# -*- coding-utf8 -*-
#############################
#    Classe MobileRobot     #
#	 autheur : YLG      	#
#	 Version : 1.1      	#
#############################

import pygame
import math
from math import *

class Point2D:
    def __init__(self, x, y):
        self.x = x
        self.y = y


class MobileRobot:
    """---------------------
         CONSTRUCTEUR   
    ---------------------"""
    def __init__(self, xA, yA, theta, scale = 1):
        """ Robot dimensions """
        self.L = 25*scale # Longueur robot
        self.l = 25*scale # longueur d'essieu arriere
        self.D = 10*scale # Diametre roue

        """ Robot's dynamics points """
        self.A = Point2D(xA, yA)  
        self.B = Point2D(xA -self.L*math.sin(theta), yA + self.L*math.cos(theta) )
        self.theta_10 = theta

        """ Robot's intputs """
        self.theta_21 = 0
        self.vA = 0

        """ Sampling rate (fps) """
        self.frameRate = 40;             # Frequence d'animation
        self.clock = pygame.time.Clock() # Objet temps
        
        """ Distances des obstacles detectes """
        self.leftDist = 400
        self.frontDist = 400
        self.rightDist = 400
        
        """ variables pour statistiques """
        self.nCollisions = 0
        self.road = 0
        self.elapsedTime = 1e-12

    """---------------------
          FONCTIONS   
    ---------------------"""

    """--- POSITION DU ROBOT ---"""
    def set_position(self, xA, yA, theta, scale = 1):
        """ Robot dimensions """
        self.L = 25*scale # Longueur robot
        self.l = 25*scale # longueur d'essieu arriere
        self.D = 10*scale # Diametre roue

        """ Robot's dynamics points """
        self.A = Point2D(xA, yA)  
        self.B = Point2D(xA -self.L*math.sin(theta), yA + self.L*math.cos(theta) )
        self.theta_10 = theta

    def get_position(self):
        return (int((self.B.x+2*self.A.x)/3), int((self.B.y+2*self.A.y)/3))
            
    """--- COMMANDE DU ROBOT ---"""
    def set_command(self, theta, speed):
        self.theta_21 = theta
        self.vA = speed

    """--- CHANGEMENT DE FRAMERATE ---"""
    def set_framerate(self, fps):
        self.frameRate = fps

    """--- CALCUL DE LA NOUVELLE POSITION ---"""
    def update(self, wait = True):

        if wait:
            self.clock.tick(self.frameRate)  # frequence d'echantillonnage fixe
        Te = 1.0/self.frameRate			     # Periode d'echantillonnage
        self.elapsedTime += Te
        
        if(abs(self.theta_21) > 1e-1):
            # rotation de centre C
            rA = self.L/math.tan(self.theta_21)
            omega_10 = self.vA/rA
            C = Point2D(self.A.x - self.L*math.cos(self.theta_10)/math.tan(self.theta_21) \
                      , self.A.y - self.L*math.sin(self.theta_10)/math.tan(self.theta_21))
            self.theta_10 += omega_10*Te
            self.A = Point2D(C.x + rA*math.cos(self.theta_10) \
                           , C.y + rA*math.sin(self.theta_10))
        else:
            # C infini : aucune rotation, translation de V_A
            self.A.x -= Te*self.vA*math.sin(self.theta_10)
            self.A.y += Te*self.vA*math.cos(self.theta_10)
            
        self.road += Te*self.vA
        
        # Calcul du nouveau point B
        temp_x = -self.L*math.sin(self.theta_10)
        temp_y = self.L*math.cos(self.theta_10)
        self.B = Point2D(self.A.x + temp_x, self.A.y + temp_y)

    """--- TRACE DU ROBOT ---"""
    def draw(self, scr, flip = False, bord = False):
        """--- Calcul des points du trace des roues arrieres ---"""
        # Centres des roues arrieres gauche et droite
        temp_x = 0.5*self.l*math.cos(self.theta_10)
        temp_y = 0.5*self.l*math.sin(self.theta_10)
        AG0 = Point2D(self.A.x - temp_x, self.A.y - temp_y) # centre gauche
        AD0 = Point2D(self.A.x + temp_x, self.A.y + temp_y) # centre droit
        
        # Roue gauche et droite
        temp_x = -0.5*self.D*math.sin(self.theta_10)
        temp_y = 0.5*self.D*math.cos(self.theta_10)
        AG1 = Point2D(AG0.x + temp_x, AG0.y + temp_y) # TL roue gauche
        AG2 = Point2D(AG0.x - temp_x, AG0.y - temp_y) # BR roue gauche
        AD1 = Point2D(AD0.x + temp_x, AD0.y + temp_y) # TL roue droite
        AD2 = Point2D(AD0.x - temp_x, AD0.y - temp_y) # BR roue droite

        """--- Calcul des points du trace de la roue avant ---"""
        temp_x = -0.5*self.D*math.sin(self.theta_10+self.theta_21)
        temp_y = 0.5*self.D*math.cos(self.theta_10+self.theta_21)
        B1 = Point2D(self.B.x + temp_x, self.B.y + temp_y)  # TL roue avant
        B2 = Point2D(self.B.x - temp_x, self.B.y - temp_y)  # BR roue avant

        height = scr.get_height()             

        """--- Trace du bord du robot (vert) ---"""
        if bord:
            if flip:
                pygame.draw.circle(scr, (255,0,0), (int((self.B.x+2*self.A.x)/3), height - int((self.B.y+2*self.A.y)/3)), int(2*self.L/3+self.D/2), 0)
            else:
                pygame.draw.circle(scr, (255,0,0), (int((self.B.x+2*self.A.x)/3), int((self.B.y+2*self.A.y)/3)), int(2*self.L/3+self.D/2), 0)
        else:
            """--- Trace des axes du robot (gris) ---"""
            if flip:
                pygame.draw.line(scr, (128,128,128), (AG0.x, height - AG0.y), (AD0.x, height - AD0.y), 2)
                pygame.draw.line(scr, (128,128,128), (self.A.x, height - self.A.y), (self.B.x, height - self.B.y), 2)          
            else:
                pygame.draw.line(scr, (128,128,128), (AG0.x, AG0.y), (AD0.x, AD0.y), 2)
                pygame.draw.line(scr, (128,128,128), (self.A.x, self.A.y), (self.B.x, self.B.y), 2)          
                     
        """--- Trace des roues du robot (noir) ---"""
        if flip:
            pygame.draw.line(scr, (0,0,0), (AG1.x, height - AG1.y), (AG2.x, height - AG2.y), 5)
            pygame.draw.line(scr, (0,0,0), (AD1.x, height - AD1.y), (AD2.x, height - AD2.y), 5)
            pygame.draw.line(scr, (0,0,0), (B1.x, height - B1.y), (B2.x, height - B2.y), 5)
        else:
            pygame.draw.line(scr, (0,0,0), (AG1.x, AG1.y), (AG2.x, AG2.y), 5)
            pygame.draw.line(scr, (0,0,0), (AD1.x, AD1.y), (AD2.x, AD2.y), 5)
            pygame.draw.line(scr, (0,0,0), (B1.x, B1.y), (B2.x, B2.y), 5)
 
    """--- LANCER DE RAYON --- """
    def radar_shot(self, scr, dst, theta, color, distMax, flip = False, trace = 0):
        
        width = scr.get_width()
        height = scr.get_height()
        coef = math.tan(theta)
        tmp_trace = (trace % 2 == 1)
         
        # Point de depart = point B + D/2 selon direction de theta 
        x = self.B.x + self.D*math.cos(theta)/2
        y = self.B.y + self.D*math.sin(theta)/2

        collised = False
        if abs(coef) <= 1:
            # delta_y < delta_x => increment en x

            inc = int(math.ceil(distMax / math.sqrt(coef**2+1)))
            sgnX = math.fabs(math.cos(theta))/math.cos(theta) 
            for i in range(inc):
                if x <= 0:
                    x = 0
                    collised = True
                    break
                elif y <= 0:
                    if flip:
                        y = height
                    else:
                        y = 0
                    collised = True
                    break
                elif y >= height:
                    if flip:
                        y = 0
                    else:
                        y = height
                    collised = True
                    break
                else:
                    try:
                        if flip:
                            col = scr.get_at((int(x),height-int(y)))
                        else:
                            col = scr.get_at((int(x), int(y)))                                       
                        if col != (255,255,255,255):
                            # On a un obstacle
                            collised = True
                            break
                        else:
                            if tmp_trace :
                                if flip:
                                    dst.set_at((int(x),height-int(y)),color)
                                else:        
                                    dst.set_at((int(x),int(y)),color)
                            x += sgnX
                            y += sgnX*coef
                    except:
                        collised = True
                        break
        else:
            # delta_y > delta_x => increment en y
            coef = 1/coef
            inc = int(math.ceil(distMax / math.sqrt(coef**2+1)))
            sgnY = math.fabs(math.sin(theta))/math.sin(theta) 
            for i in range(inc):
                if x <= 0:
                    x = 0
                    collised = True
                    break
                elif y <= 0:
                    if flip:
                        y = height
                    else:
                        y = 0
                    collised = True
                    break
                elif x >= width:
                    x = width
                    collised = True
                    break
                else:
                    try:
                        if flip:
                            col = scr.get_at((int(x),height-int(y)))
                        else:
                            col = scr.get_at((int(x), int(y)))                                       
                        if col != (255,255,255,255):
                            # On a un obstacle
                            collised = True
                            break
                        else:
                            if tmp_trace:
                                if flip:
                                    dst.set_at((int(x),height-int(y)),color)
                                else:        
                                    dst.set_at((int(x),int(y)),color)
                            y += sgnY
                            x += sgnY*coef
                    except:
                        # On a un obstacle
                        collised = True
                        break
        
        if collised:
            return math.sqrt((self.B.x-x)**2+(self.B.y-y)**2) # on retourne la distance a l'obstacle
        else:
            return width+height #distance "infinie"
        
    def my_range(self, start, end, step):
        while start <= end:
            yield start
            start += step

    """--- SIMULATION DE 3 TELEMETRES A ULTRASONS ---"""
    #   _________________________              __________________________
    #  /                         \            /                          \
    #  |  Parametres pour chaque |            |  Orientation globale des | 
    #  |  capteur simule :       |            |  capteurs :              |   
    #  \_________________________/            \__________________________/     
    #                                                 
    #    \ zone de detection /                   Directions des 3 capteurs             
    #     \<--------------->/                               ,
    #      \               /  ,                _           /|\           _
    #       \      |      /  /|\              |\_           |           _/|
    #        \alpha|     /    |                  \_   beta  |         _/
    #         \  ,-|-,  /     |                    \_   ,---|---,   _/
    #          \/  |  \/      | distMax              \_/    |    \_/ 
    #           \  |  /       |                        \_   |   _/
    #            \ | /        |                         _\,-|-./_
    #             \|/        \|/                       /         \ 
    #         _____v_____     '                       |   ROBOT   |
    #     ___/           \___                         |           |
    #    /       ROBOT       \                        
    #
    #    avec nShot lancers dans
    #         chaque zone
    #
    def us_sensor(self, scr, dst, nShot, distMax = -1, alpha = pi/6, beta = pi/3, flip = False, trace = 0):
        thetaMin = 0
        dist = 0
        width, height = scr.get_width(), scr.get_height()
        if distMax < 0:
            distMax = width + height
            
        # Right Scan
        self.rightDist = width + height
        for theta in self.my_range(pi/2-beta-alpha, pi/2-beta+alpha, 2*alpha/nShot):
            dist = self.radar_shot(scr, dst, theta+self.theta_10, (0,255,255), distMax, flip, trace)
            if dist < self.rightDist:
                self.rightDist = dist
                thetaMin = theta
        if trace >= 2 and self.rightDist < distMax:
            self.radar_shot(scr, dst, thetaMin+self.theta_10, (0,255,255), dist, flip, 1)
            
        # Front Scan
        self.frontDist = width + height
        for theta in self.my_range(pi/2-alpha, pi/2+alpha, 2*alpha/nShot):
            dist = self.radar_shot(scr, dst, theta+self.theta_10, (255,255,0), distMax, flip, trace)
            if dist < self.frontDist:
                self.frontDist = dist
                thetaMin = theta
        if trace >= 2 and self.frontDist < distMax:
            self.radar_shot(scr, dst, thetaMin+self.theta_10, (255,255,0), dist, flip, 1)
                
        # Left Scan
        self.leftDist = width + height
        for theta in self.my_range(pi/2+beta-alpha, pi/2+beta+alpha, 2*alpha/nShot):
            dist = self.radar_shot(scr, dst, theta+self.theta_10, (255,0,255), distMax, flip, trace)
            if dist < self.leftDist:
                self.leftDist = dist
                thetaMin = theta
        if trace >= 2 and self.leftDist < distMax:
            self.radar_shot(scr, dst, thetaMin+self.theta_10, (255,0,255), dist, flip, 1)
            
    """--- SIMULATION D'UN CAPTEUR DE LUMINOSITE ---"""
    #  Renvoi la luminosite moyenne d'un disque de centre B et de diametre d pixels
    def light_sensor(self, scr, d = 3, flip = False):
        x = self.B.x
        y = self.B.y
        light = 0
        surf = 0
        height = scr.get_height()
        if d <= 0:
            dd = 1
        else:
            dd = d - (d+1)%2
        step = (dd-1)//2
        for i in range(-step, step+1):
            for j in range(-step, step+1):
                if (i**2+j**2) <= (d/2)**2:
                    try:
                        if flip:
                            col = scr.get_at((int(x+i),height-int(y+j)))
                        else:
                            col = scr.get_at((int(x+i),int(y+j)))
                        for c in range(3):
                            light += col[c]
                        surf += 3
                    except:
                        pass
        if surf > 0:
            return light/surf
        else:
            return -1

    """--- TEST DE COLLISION DU ROBOT ---"""
    def collision(self, screen, flip=False):
        # scan sur les bords du robot pour savoir si un pixel est allume avant de le tracer
        R = int(2*self.L/3+self.D/2) #rayon du cercle definissant les bords du robot
        tmpC = Point2D(int((self.B.x+2*self.A.x)/3), int((self.B.y+2*self.A.y)/3)) #centre du cercle
        height = screen.get_height()
        for theta in self.my_range(0,2*pi,pi/200): # boucle de 0 a 360 sur ce cercle
            try:
                if flip:
                    if screen.get_at((int(tmpC.x + R*math.cos(theta)), int(height - (tmpC.y + R*math.sin(theta))))) != (255,255,255,255):
                        self.nCollisions += 1 # si pixel allume = collision
                        return True           # on arrete aussitot en retournant "vrai"
                else:
                    if screen.get_at((int(tmpC.x + R*math.cos(theta)), int(tmpC.y + R*math.sin(theta)))) != (255,255,255,255): 
                        self.nCollisions += 1 # si pixel allume = collision
                        return True           # on arrete aussitot en retournant "vrai"
            except: 
                # arrive quand pixel en dehors de l'ecran
                self.nCollisions += 1   # equivaut a une collision
                return True             # on arrete aussitot en retournant "vrai"
        return False    
